@@ -1,4 +1,4 @@
-/* See LICENSE file for license and copyright information */
+/* SPDX-License-Identifier: Zlib */
 
 #include <girara/statusbar.h>
 #include <girara/session.h>
@@ -103,8 +103,8 @@ cb_view_hadjustment_value_changed(GtkAdjustment* adjustment, gpointer data)
 
   update_visible_pages(zathura);
 
-  double position_x = zathura_adjustment_get_ratio(adjustment);
-  double position_y = zathura_document_get_position_y(zathura->document);
+  const double position_x = zathura_adjustment_get_ratio(adjustment);
+  const double position_y = zathura_document_get_position_y(zathura->document);
   unsigned int page_id = position_to_page_number(zathura->document, position_x, position_y);
 
   zathura_document_set_position_x(zathura->document, position_x);
@@ -129,9 +129,9 @@ cb_view_vadjustment_value_changed(GtkAdjustment* adjustment, gpointer data)
 
   update_visible_pages(zathura);
 
-  double position_x = zathura_document_get_position_x(zathura->document);
-  double position_y = zathura_adjustment_get_ratio(adjustment);
-  unsigned int page_id = position_to_page_number(zathura->document, position_x, position_y);
+  const double position_x = zathura_document_get_position_x(zathura->document);
+  const double position_y = zathura_adjustment_get_ratio(adjustment);
+  const unsigned int page_id = position_to_page_number(zathura->document, position_x, position_y);
 
   zathura_document_set_position_x(zathura->document, position_x);
   zathura_document_set_position_y(zathura->document, position_y);
@@ -158,7 +158,7 @@ cb_view_adjustment_changed(GtkAdjustment* adjustment, zathura_t* zathura,
   }
 
   /* Save the viewport size */
-  unsigned int size = (unsigned int)floor(gtk_adjustment_get_page_size(adjustment));
+  const unsigned int size = floor(gtk_adjustment_get_page_size(adjustment));
   if (width == true) {
     zathura_document_set_viewport_width(zathura->document, size);
   } else {
@@ -204,11 +204,20 @@ cb_refresh_view(GtkWidget* GIRARA_UNUSED(view), gpointer data)
     return;
   }
 
+  if (zathura->pages != NULL && zathura->pages[page_id] != NULL) {
+	  ZathuraPage* page_widget = ZATHURA_PAGE(zathura->pages[page_id]);
+	  if (page_widget != NULL) {
+		  if (zathura_page_widget_have_surface(page_widget)) {
+			  document_predecessor_free(zathura);
+		  }
+	  }
+  }
+
   GtkAdjustment* vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
   GtkAdjustment* hadj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
 
-  double position_x = zathura_document_get_position_x(zathura->document);
-  double position_y = zathura_document_get_position_y(zathura->document);
+  const double position_x = zathura_document_get_position_x(zathura->document);
+  const double position_y = zathura_document_get_position_y(zathura->document);
 
   zathura_adjustment_set_value_from_ratio(vadj, position_y);
   zathura_adjustment_set_value_from_ratio(hadj, position_x);
@@ -217,7 +226,84 @@ cb_refresh_view(GtkWidget* GIRARA_UNUSED(view), gpointer data)
 }
 
 void
-cb_page_layout_value_changed(girara_session_t* session, const char* name, girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
+cb_monitors_changed(GdkScreen* screen, gpointer data)
+{
+  girara_debug("signal received");
+
+  zathura_t* zathura = data;
+  if (screen == NULL || zathura == NULL) {
+    return;
+  }
+
+  zathura_update_view_ppi(zathura);
+}
+
+void
+cb_widget_screen_changed(GtkWidget* widget, GdkScreen* previous_screen, gpointer data)
+{
+  girara_debug("called");
+
+  zathura_t* zathura = data;
+  if (widget == NULL || zathura == NULL) {
+    return;
+  }
+
+  /* disconnect previous screen handler if present */
+  if (previous_screen != NULL && zathura->signals.monitors_changed_handler > 0) {
+    g_signal_handler_disconnect(previous_screen, zathura->signals.monitors_changed_handler);
+    zathura->signals.monitors_changed_handler = 0;
+  }
+
+  if (gtk_widget_has_screen(widget)) {
+    GdkScreen* screen = gtk_widget_get_screen(widget);
+
+    /* connect to new screen */
+    zathura->signals.monitors_changed_handler = g_signal_connect(G_OBJECT(screen),
+        "monitors-changed", G_CALLBACK(cb_monitors_changed), zathura);
+  }
+
+  zathura_update_view_ppi(zathura);
+}
+
+gboolean
+cb_widget_configured(GtkWidget* UNUSED(widget), GdkEvent* UNUSED(event), gpointer data)
+{
+  zathura_t* zathura = data;
+  if (zathura == NULL) {
+    return false;
+  }
+
+  zathura_update_view_ppi(zathura);
+
+  return false;
+}
+
+void
+cb_scale_factor(GObject* object, GParamSpec* UNUSED(pspec), gpointer data)
+{
+  zathura_t* zathura = data;
+  if (zathura == NULL || zathura->document == NULL) {
+    return;
+  }
+
+  int new_factor = gtk_widget_get_scale_factor(GTK_WIDGET(object));
+  if (new_factor == 0) {
+    girara_debug("Ignoring new device scale factor = 0");
+    return;
+  }
+
+  zathura_device_factors_t current = zathura_document_get_device_factors(zathura->document);
+  if (fabs(new_factor - current.x) >= DBL_EPSILON ||
+      fabs(new_factor - current.y) >= DBL_EPSILON) {
+    zathura_document_set_device_factors(zathura->document, new_factor, new_factor);
+    girara_debug("New device scale factor: %d", new_factor);
+    zathura_update_view_ppi(zathura);
+    render_all(zathura);
+  }
+}
+
+void
+cb_page_layout_value_changed(girara_session_t* session, const char* name, girara_setting_type_t UNUSED(type), const void* value, void* UNUSED(data))
 {
   g_return_if_fail(value != NULL);
   g_return_if_fail(session != NULL);
@@ -226,7 +312,7 @@ cb_page_layout_value_changed(girara_session_t* session, const char* name, girara
 
   /* pages-per-row must not be 0 */
   if (g_strcmp0(name, "pages-per-row") == 0) {
-    unsigned int pages_per_row = *((unsigned int*) value);
+    unsigned int pages_per_row = *((const unsigned int*) value);
     if (pages_per_row == 0) {
       pages_per_row = 1;
       girara_setting_set(session, name, &pages_per_row);
@@ -254,7 +340,10 @@ cb_page_layout_value_changed(girara_session_t* session, const char* name, girara
   unsigned int page_padding = 1;
   girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
 
-  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column);
+  bool page_right_to_left = false;
+  girara_setting_get(zathura->ui.session, "page-right-to-left", &page_right_to_left);
+
+  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column, page_right_to_left);
   zathura_document_set_page_layout(zathura->document, page_padding, pages_per_row, first_page_column);
 }
 
@@ -274,7 +363,7 @@ cb_index_row_activated(GtkTreeView* tree_view, GtkTreePath* path,
 
   if(gtk_tree_model_get_iter(model, &iter, path)) {
     zathura_index_element_t* index_element;
-    gtk_tree_model_get(model, &iter, 2, &index_element, -1);
+    gtk_tree_model_get(model, &iter, 3, &index_element, -1);
 
     if (index_element == NULL) {
       return;
@@ -290,10 +379,11 @@ cb_index_row_activated(GtkTreeView* tree_view, GtkTreePath* path,
 typedef enum zathura_link_action_e
 {
   ZATHURA_LINK_ACTION_FOLLOW,
+  ZATHURA_LINK_ACTION_COPY,
   ZATHURA_LINK_ACTION_DISPLAY
 } zathura_link_action_t;
 
-static bool
+static gboolean
 handle_link(GtkEntry* entry, girara_session_t* session,
             zathura_link_action_t action)
 {
@@ -301,59 +391,64 @@ handle_link(GtkEntry* entry, girara_session_t* session,
   g_return_val_if_fail(session->global.data != NULL, FALSE);
 
   zathura_t* zathura = session->global.data;
-  bool eval = true;
+  gboolean eval = TRUE;
 
   char* input = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
   if (input == NULL || strlen(input) == 0) {
-    eval = false;
+    eval = FALSE;
   }
 
   int index = 0;
-  if (eval == true) {
+  if (eval == TRUE) {
     index = atoi(input);
     if (index == 0 && g_strcmp0(input, "0") != 0) {
       girara_notify(session, GIRARA_WARNING, _("Invalid input '%s' given."), input);
-      eval = false;
+      eval = FALSE;
     }
     index = index - 1;
   }
 
-  /* set pages to draw links */
-  bool invalid_index = true;
+  /* find the link*/
+  zathura_link_t* link = NULL;
   unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
   for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
     zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
-    if (page == NULL || zathura_page_get_visibility(page) == false) {
+    if (page == NULL || zathura_page_get_visibility(page) == false || eval == false) {
       continue;
     }
-
     GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
-    g_object_set(G_OBJECT(page_widget), "draw-links", FALSE, NULL);
+    link = zathura_page_widget_link_get(ZATHURA_PAGE(page_widget), index);
+    if (link != NULL) {
+      break;
+    }
+  }
 
-    if (eval == true) {
-      zathura_link_t* link = zathura_page_widget_link_get(ZATHURA_PAGE(page_widget), index);
-
-      if (link != NULL) {
-        invalid_index = false;
-        switch (action) {
-          case ZATHURA_LINK_ACTION_FOLLOW:
-            zathura_link_evaluate(zathura, link);
-            break;
-          case ZATHURA_LINK_ACTION_DISPLAY:
-            zathura_link_display(zathura, link);
-            break;
+  if (eval == TRUE && link == NULL) {
+    girara_notify(session, GIRARA_WARNING, _("Invalid index '%s' given."), input);
+  } else {
+    switch (action) {
+      case ZATHURA_LINK_ACTION_FOLLOW:
+        zathura_link_evaluate(zathura, link);
+        break;
+      case ZATHURA_LINK_ACTION_DISPLAY:
+        zathura_link_display(zathura, link);
+        break;
+      case ZATHURA_LINK_ACTION_COPY: {
+        GdkAtom* selection = get_selection(zathura);
+        if (selection == NULL) {
+          break;
         }
+
+        zathura_link_copy(zathura, link, selection);
+        g_free(selection);
+        break;
       }
     }
   }
 
-  if (eval == true && invalid_index == true) {
-    girara_notify(session, GIRARA_WARNING, _("Invalid index '%s' given."), input);
-  }
-
   g_free(input);
 
-  return (eval == TRUE) ? TRUE : FALSE;
+  return eval;
 }
 
 gboolean
@@ -370,6 +465,13 @@ cb_sc_display_link(GtkEntry* entry, void* data)
   return handle_link(entry, session, ZATHURA_LINK_ACTION_DISPLAY);
 }
 
+gboolean
+cb_sc_copy_link(GtkEntry* entry, void* data)
+{
+  girara_session_t* session = data;
+  return handle_link(entry, session, ZATHURA_LINK_ACTION_COPY);
+}
+
 static gboolean
 file_monitor_reload(void* data)
 {
@@ -437,7 +539,7 @@ cb_password_dialog(GtkEntry* entry, void* data)
 
   /* try to open document again */
   if (document_open(dialog->zathura, dialog->path, dialog->uri, input,
-                    ZATHURA_PAGE_NUMBER_UNSPECIFIED) == false) {
+                    ZATHURA_PAGE_NUMBER_UNSPECIFIED, NULL) == false) {
     gdk_threads_add_idle(password_dialog, dialog);
   } else {
     g_free(dialog->path);
@@ -479,7 +581,7 @@ cb_view_resized(GtkWidget* UNUSED(widget), GtkAllocation* UNUSED(allocation), za
 
 void
 cb_setting_recolor_change(girara_session_t* session, const char* name,
-                          girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
+                          girara_setting_type_t UNUSED(type), const void* value, void* UNUSED(data))
 {
   g_return_if_fail(value != NULL);
   g_return_if_fail(session != NULL);
@@ -487,7 +589,7 @@ cb_setting_recolor_change(girara_session_t* session, const char* name,
   g_return_if_fail(name != NULL);
   zathura_t* zathura = session->global.data;
 
-  const bool bool_value = *((bool*) value);
+  const bool bool_value = *((const bool*) value);
 
   if (zathura->sync.render_thread != NULL && zathura_renderer_recolor_enabled(zathura->sync.render_thread) != bool_value) {
      zathura_renderer_enable_recolor(zathura->sync.render_thread, bool_value);
@@ -497,7 +599,7 @@ cb_setting_recolor_change(girara_session_t* session, const char* name,
 
 void
 cb_setting_recolor_keep_hue_change(girara_session_t* session, const char* name,
-                                   girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
+                                   girara_setting_type_t UNUSED(type), const void* value, void* UNUSED(data))
 {
   g_return_if_fail(value != NULL);
   g_return_if_fail(session != NULL);
@@ -505,7 +607,7 @@ cb_setting_recolor_keep_hue_change(girara_session_t* session, const char* name,
   g_return_if_fail(name != NULL);
   zathura_t* zathura = session->global.data;
 
-  const bool bool_value = *((bool*) value);
+  const bool bool_value = *((const bool*) value);
 
   if (zathura->sync.render_thread != NULL && zathura_renderer_recolor_hue_enabled(zathura->sync.render_thread) != bool_value) {
      zathura_renderer_enable_recolor_hue(zathura->sync.render_thread, bool_value);
@@ -515,7 +617,7 @@ cb_setting_recolor_keep_hue_change(girara_session_t* session, const char* name,
 
 void
 cb_setting_recolor_keep_reverse_video_change(girara_session_t* session, const char* name,
-                                   girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
+                                   girara_setting_type_t UNUSED(type), const void* value, void* UNUSED(data))
 {
   g_return_if_fail(value != NULL);
   g_return_if_fail(session != NULL);
@@ -523,7 +625,7 @@ cb_setting_recolor_keep_reverse_video_change(girara_session_t* session, const ch
   g_return_if_fail(name != NULL);
   zathura_t* zathura = session->global.data;
 
-  const bool bool_value = *((bool*) value);
+  const bool bool_value = *((const bool*) value);
 
   if (zathura->sync.render_thread != NULL && zathura_renderer_recolor_reverse_video_enabled(zathura->sync.render_thread) != bool_value) {
      zathura_renderer_enable_recolor_reverse_video(zathura->sync.render_thread, bool_value);
@@ -573,27 +675,28 @@ cb_page_widget_text_selected(ZathuraPage* page, const char* text, void* data)
   }
 
   GdkAtom* selection = get_selection(zathura);
+  if (selection == NULL) {
+    return;
+  }
 
   /* copy to clipboard */
-  if (selection != NULL) {
-    gtk_clipboard_set_text(gtk_clipboard_get(*selection), text, -1);
+  gtk_clipboard_set_text(gtk_clipboard_get(*selection), text, -1);
 
-    bool notification = true;
-    girara_setting_get(zathura->ui.session, "selection-notification", &notification);
+  bool notification = true;
+  girara_setting_get(zathura->ui.session, "selection-notification", &notification);
 
-    if (notification == true) {
-      char* target = NULL;
-      girara_setting_get(zathura->ui.session, "selection-clipboard", &target);
+  if (notification == true) {
+    char* target = NULL;
+    girara_setting_get(zathura->ui.session, "selection-clipboard", &target);
 
-      char* stripped_text = g_strdelimit(g_strdup(text), "\n\t\r\n", ' ');
-      char* escaped_text = g_markup_printf_escaped(
-          _("Copied selected text to selection %s: %s"), target, stripped_text);
-      g_free(target);
-      g_free(stripped_text);
+    char* stripped_text = g_strdelimit(g_strdup(text), "\n\t\r\n", ' ');
+    char* escaped_text = g_markup_printf_escaped(
+        _("Copied selected text to selection %s: %s"), target, stripped_text);
+    g_free(target);
+    g_free(stripped_text);
 
-      girara_notify(zathura->ui.session, GIRARA_INFO, "%s", escaped_text);
-      g_free(escaped_text);
-    }
+    girara_notify(zathura->ui.session, GIRARA_INFO, "%s", escaped_text);
+    g_free(escaped_text);
   }
 
   g_free(selection);
@@ -646,41 +749,102 @@ cb_page_widget_link(ZathuraPage* page, void* data)
 
 void
 cb_page_widget_scaled_button_release(ZathuraPage* page_widget, GdkEventButton* event,
-    void* data)
+                                     void* data)
+{
+  zathura_t* zathura = data;
+  zathura_page_t* page = zathura_page_widget_get_page(page_widget);
+
+  if (event->button != GDK_BUTTON_PRIMARY) {
+    return;
+  }
+
+  /* set page number (but don't scroll there. it was clicked on, so it's visible) */
+  zathura_document_set_current_page_number(zathura->document, zathura_page_get_index(page));
+  refresh_view(zathura);
+
+  if (event->state & zathura->global.synctex_edit_modmask) {
+    bool synctex = false;
+    girara_setting_get(zathura->ui.session, "synctex", &synctex);
+    if (synctex == false) {
+      return;
+    }
+
+    if (zathura->dbus != NULL) {
+      zathura_dbus_edit(zathura->dbus, zathura_page_get_index(page), event->x, event->y);
+    }
+
+    char* editor = NULL;
+    girara_setting_get(zathura->ui.session, "synctex-editor-command", &editor);
+    if (editor == NULL || *editor == '\0') {
+      girara_debug("No SyncTeX editor specified.");
+      g_free(editor);
+      return;
+    }
+
+    synctex_edit(editor, page, event->x, event->y);
+    g_free(editor);
+  }
+}
+
+void
+cb_window_update_icon(ZathuraRenderRequest* GIRARA_UNUSED(request), cairo_surface_t* surface, void* data)
 {
   zathura_t* zathura = data;
 
-  zathura_page_t* page = zathura_page_widget_get_page(page_widget);
-
-  /* set page number (but don't scroll there. it was clicked on, so it's visible) */
-  if (event->button == GDK_BUTTON_PRIMARY) {
-    zathura_document_set_current_page_number(zathura->document, zathura_page_get_index(page));
-    refresh_view(zathura);
-  }
-
-  if (event->button != 1 || !(event->state & GDK_CONTROL_MASK)) {
+  girara_debug("updating window icon");
+  GdkPixbuf* pixbuf = gdk_pixbuf_get_from_surface(surface, 0, 0, cairo_image_surface_get_width(surface), cairo_image_surface_get_height(surface));
+  if (pixbuf == NULL) {
+    girara_error("Unable to convert cairo surface to Gdk Pixbuf.");
     return;
   }
 
-  bool synctex = false;
-  girara_setting_get(zathura->ui.session, "synctex", &synctex);
-
-  if (synctex == false) {
-    return;
-  }
-
-  if (zathura->dbus != NULL) {
-    zathura_dbus_edit(zathura->dbus, zathura_page_get_index(page), event->x, event->y);
-  }
-
-  char* editor = NULL;
-  girara_setting_get(zathura->ui.session, "synctex-editor-command", &editor);
-  if (editor == NULL || *editor == '\0') {
-    g_free(editor);
-    return;
-  }
-
-  synctex_edit(editor, page, event->x, event->y);
-  g_free(editor);
+  gtk_window_set_icon(GTK_WINDOW(zathura->ui.session->gtk.window), pixbuf);
+  g_object_unref(pixbuf);
 }
 
+void
+cb_gesture_zoom_begin(GtkGesture* UNUSED(self), GdkEventSequence* UNUSED(sequence), void* data)
+{
+  zathura_t* zathura = data;
+  if (zathura == NULL || zathura->document == NULL) {
+    return;
+  }
+  const double current_zoom     = zathura_document_get_zoom(zathura->document);
+  zathura->gesture.initial_zoom = current_zoom;
+}
+
+void
+cb_gesture_zoom_scale_changed(GtkGestureZoom* UNUSED(self), gdouble scale, void* data)
+{
+  zathura_t* zathura = data;
+  if (zathura == NULL || zathura->document == NULL) {
+    return;
+  }
+
+  const double next_zoom = zathura->gesture.initial_zoom * scale;
+  const double corrected_zoom = zathura_correct_zoom_value(zathura->ui.session, next_zoom);
+  zathura_document_set_zoom(zathura->document, corrected_zoom);
+  render_all(zathura);
+  refresh_view(zathura);
+}
+
+void cb_hide_links(GtkWidget* widget, gpointer data) {
+  g_return_if_fail(widget != NULL);
+  g_return_if_fail(data != NULL);
+
+  /* disconnect from signal */
+  gulong handler_id = GPOINTER_TO_UINT(g_object_steal_data(G_OBJECT(widget), "handler_id"));
+  g_signal_handler_disconnect(G_OBJECT(widget), handler_id);
+
+  zathura_t* zathura           = data;
+  unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
+  for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
+    zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
+    if (page == NULL) {
+      continue;
+    }
+
+    GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
+    g_object_set(G_OBJECT(page_widget), "draw-links", FALSE, NULL);
+  }
+}
