@@ -1,4 +1,4 @@
-/* See LICENSE file for license and copyright information */
+/* SPDX-License-Identifier: Zlib */
 
 #ifndef ZATHURA_DATABASE_SQLITE_H
 #define ZATHURA_DATABASE_SQLITE_H
@@ -6,7 +6,7 @@
 #include "database.h"
 
 #define ZATHURA_TYPE_SQLDATABASE \
-  (zathura_sqldatabase_get_type ())
+  (zathura_sqldatabase_get_type())
 #define ZATHURA_SQLDATABASE(obj) \
   (G_TYPE_CHECK_INSTANCE_CAST ((obj), ZATHURA_TYPE_SQLDATABASE, ZathuraSQLDatabase))
 #define ZATHURA_IS_SQLDATABASE(obj) \
@@ -41,4 +41,13 @@ GType zathura_sqldatabase_get_type(void) G_GNUC_CONST;
  */
 zathura_database_t* zathura_sqldatabase_new(const char* path);
 
+/**
+ * Initialize database system from an old plain database.
+ *
+ * @param sqlite_path Path to the sqlite database.
+ * @param plain_dir Path to the old plain database.
+ * @return A valid zathura_database_t instance or NULL on failure
+ */
+zathura_database_t* zathura_sqldatabase_new_from_plain(const char* sqlite_path, const char* plain_dir);
+
 #endif
