@@ -1,5 +1,6 @@
 /* SPDX-License-Identifier: Zlib */
 
+#include <girara/log.h>
 #include <girara/session.h>
 #include <girara/settings.h>
 #include <girara/datastructures.h>
@@ -10,6 +11,7 @@
 
 #include "callbacks.h"
 #include "shortcuts.h"
+#include "dbus-interface.h"
 #include "document.h"
 #include "zathura.h"
 #include "render.h"
@@ -18,9 +20,10 @@
 #include "print.h"
 #include "page-widget.h"
 #include "adjustment.h"
+#include "database.h"
 #include <math.h>
 
-/* Helper function; see sc_display_link and sc_follow. */
+/* Helper function for highlighting the links */
 static bool
 draw_links(zathura_t* zathura)
 {
@@ -54,6 +57,30 @@ draw_links(zathura_t* zathura)
   return show_links;
 }
 
+/* Common code for sc_follow, sc_display_link and sc_copy_link */
+static bool
+link_shortcuts(zathura_t* zathura, girara_callback_inputbar_activate_t callback,
+		       const char* text)
+{
+  if (zathura->document == NULL || zathura->ui.session == NULL) {
+    return false;
+  }
+
+  bool show_links = draw_links(zathura);
+
+  /* ask for input */
+  if (show_links == true) {
+    GtkWidget *inputbar = zathura->ui.session->gtk.inputbar;
+    gulong handler_id = g_signal_connect(inputbar, "hide", G_CALLBACK(cb_hide_links), zathura);
+    g_object_set_data(G_OBJECT(inputbar), "handler_id", GUINT_TO_POINTER(handler_id));
+
+    zathura_document_set_adjust_mode(zathura->document, ZATHURA_ADJUST_INPUTBAR);
+    girara_dialog(zathura->ui.session, text, FALSE, NULL, callback, zathura->ui.session);
+  }
+
+  return false;
+}
+
 bool
 sc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument),
          girara_event_t* UNUSED(event), unsigned int UNUSED(t))
@@ -124,6 +151,34 @@ sc_change_mode(girara_session_t* session, girara_argument_t* argument,
   return false;
 }
 
+bool
+sc_cycle_first_column(girara_session_t* session, girara_argument_t* UNUSED(argument),
+                      girara_event_t* UNUSED(event), unsigned int t)
+{
+  g_return_val_if_fail(session != NULL, false);
+  g_return_val_if_fail(session->global.data != NULL, false);
+  zathura_t* zathura = session->global.data;
+
+  if (zathura->document == NULL) {
+    girara_notify(session, GIRARA_WARNING, _("No document opened."));
+    return false;
+  }
+
+  int pages_per_row = 1;
+  girara_setting_get(session, "pages-per-row", &pages_per_row);
+  char* first_page_column_list = NULL;
+  girara_setting_get(session, "first-page-column", &first_page_column_list);
+
+  if (t == 0) t = 1;
+  char* new_column_list = increment_first_page_column(first_page_column_list, pages_per_row, t);
+  g_free(first_page_column_list);
+
+  girara_setting_set(session, "first-page-column", new_column_list);
+  g_free(new_column_list);
+
+  return true;
+}
+
 bool
 sc_display_link(girara_session_t* session, girara_argument_t* UNUSED(argument),
                 girara_event_t* UNUSED(event), unsigned int UNUSED(t))
@@ -132,21 +187,7 @@ sc_display_link(girara_session_t* session, girara_argument_t* UNUSED(argument),
   g_return_val_if_fail(session->global.data != NULL, false);
   zathura_t* zathura = session->global.data;
 
-  if (zathura->document == NULL || zathura->ui.session == NULL) {
-    return false;
-  }
-
-  bool show_links = draw_links(zathura);
-
-  /* ask for input */
-  if (show_links) {
-    zathura_document_set_adjust_mode(zathura->document, ZATHURA_ADJUST_INPUTBAR);
-    girara_dialog(zathura->ui.session, "Display link:", FALSE, NULL,
-        cb_sc_display_link,
-        zathura->ui.session);
-  }
-
-  return false;
+  return link_shortcuts(zathura, cb_sc_display_link, "Display Link: ");
 }
 
 bool
@@ -157,21 +198,32 @@ sc_copy_link(girara_session_t* session, girara_argument_t* UNUSED(argument),
   g_return_val_if_fail(session->global.data != NULL, false);
   zathura_t* zathura = session->global.data;
 
-  if (zathura->document == NULL || zathura->ui.session == NULL) {
-    return false;
-  }
+  return link_shortcuts(zathura, cb_sc_copy_link, "Copy Link: ");
+}
 
-  bool show_links = draw_links(zathura);
+bool
+sc_copy_filepath(girara_session_t* session, girara_argument_t*  UNUSED(argument), girara_event_t* UNUSED(event), unsigned int UNUSED(t))
+{
+    g_return_val_if_fail(session != NULL, false);
+    g_return_val_if_fail(session->global.data != NULL, false);
+    zathura_t* zathura = session->global.data;
 
-  /* ask for input */
-  if (show_links) {
-    zathura_document_set_adjust_mode(zathura->document, ZATHURA_ADJUST_INPUTBAR);
-    girara_dialog(zathura->ui.session, "Copy link:", FALSE, NULL,
-        cb_sc_copy_link,
-        zathura->ui.session);
-  }
+    GdkAtom* selection = get_selection(zathura);
+    if (selection == NULL) {
+        return false;
+    }
 
-  return false;
+    const char* file_path = zathura_document_get_path(zathura->document);
+    if (file_path == NULL) {
+        girara_debug("Could not get file path for copying");
+        return false;
+    }
+
+    girara_debug("Copying file path to clipboard");
+    gtk_clipboard_set_text(gtk_clipboard_get(*selection), file_path, -1);
+
+    g_free(selection);
+    return true;
 }
 
 bool
@@ -246,19 +298,7 @@ sc_follow(girara_session_t* session, girara_argument_t* UNUSED(argument),
   g_return_val_if_fail(session->global.data != NULL, false);
   zathura_t* zathura = session->global.data;
 
-  if (zathura->document == NULL || zathura->ui.session == NULL) {
-    return false;
-  }
-
-  bool show_links = draw_links(zathura);
-
-  /* ask for input */
-  if (show_links == true) {
-    zathura_document_set_adjust_mode(zathura->document, ZATHURA_ADJUST_INPUTBAR);
-    girara_dialog(zathura->ui.session, "Follow link:", FALSE, NULL, cb_sc_follow, zathura->ui.session);
-  }
-
-  return false;
+  return link_shortcuts(zathura, cb_sc_follow, "Follow Link: ");
 }
 
 bool
@@ -470,13 +510,16 @@ sc_reload(girara_session_t* session, girara_argument_t* UNUSED(argument),
     return false;
   }
 
+  /* Get file info (zoom, current page, etc.) */
+  zathura_fileinfo_t file_info = zathura_get_fileinfo(zathura);
+
   /* close current document */
   document_close(zathura, true);
 
-  /* reopen document */
+  /* reopen document with old file info */
   document_open(
     zathura, zathura_filemonitor_get_filepath(zathura->file_monitor.monitor),
-    NULL, zathura->file_monitor.password, ZATHURA_PAGE_NUMBER_UNSPECIFIED);
+    NULL, zathura->file_monitor.password, file_info.current_page, &file_info );
 
   return false;
 }
@@ -503,7 +546,7 @@ sc_rotate(girara_session_t* session, girara_argument_t* argument,
   zathura_document_set_rotation(zathura->document, (rotation + angle * t) % 360);
 
   /* update scale */
-  girara_argument_t new_argument = { zathura_document_get_adjust_mode(zathura->document), NULL };
+  girara_argument_t new_argument = {.n = zathura_document_get_adjust_mode(zathura->document), .data = NULL};
   sc_adjust_window(zathura->ui.session, &new_argument, NULL, 0);
 
   /* render all pages again */
@@ -1021,17 +1064,28 @@ sc_navigate_index(girara_session_t* session, girara_argument_t* argument,
 
   GtkTreeView *tree_view = gtk_container_get_children(GTK_CONTAINER(zathura->ui.index))->data;
   GtkTreePath *path;
+  GtkTreePath *start_path;
+  GtkTreePath *end_path;
 
   gtk_tree_view_get_cursor(tree_view, &path, NULL);
   if (path == NULL) {
     return false;
   }
 
+  if (gtk_tree_view_get_visible_range(tree_view, &start_path, &end_path) != TRUE)
+  {
+    girara_error("Cannot get visible range for index");
+    gtk_tree_path_free(start_path);
+    gtk_tree_path_free(end_path);
+    return false;
+  }
+
   GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
   GtkTreeIter   iter;
   GtkTreeIter   child_iter;
 
   gboolean is_valid_path = TRUE;
+  gboolean need_to_scroll = FALSE;
 
   switch(argument->n) {
     case TOP:
@@ -1087,6 +1141,18 @@ sc_navigate_index(girara_session_t* session, girara_argument_t* argument,
                 && gtk_tree_path_up(path));
       }
       break;
+    case HALF_UP:
+      gtk_tree_path_free(path);
+      gtk_tree_path_free(end_path);
+      path = start_path;
+      need_to_scroll = TRUE;
+      break;
+    case HALF_DOWN:
+      gtk_tree_path_free(path);
+      gtk_tree_path_free(start_path);
+      path = end_path;
+      need_to_scroll = TRUE;
+      break;
     case EXPAND:
       if (gtk_tree_view_expand_row(tree_view, path, FALSE)) {
         gtk_tree_path_down(path);
@@ -1119,6 +1185,9 @@ sc_navigate_index(girara_session_t* session, girara_argument_t* argument,
 
   if (is_valid_path == TRUE) {
     gtk_tree_view_set_cursor(tree_view, path, NULL, FALSE);
+    if (need_to_scroll == TRUE) {
+      gtk_tree_view_scroll_to_cell(tree_view, path, NULL, TRUE, 0.5, 0.0);
+    }
   }
 
   gtk_tree_path_free(path);
@@ -1161,7 +1230,7 @@ sc_toggle_index(girara_session_t* session, girara_argument_t* UNUSED(argument),
       goto error_free;
     }
 
-    model = GTK_TREE_MODEL(gtk_tree_store_new(3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER));
+    model = GTK_TREE_MODEL(gtk_tree_store_new(4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER));
     if (model == NULL) {
       goto error_free;
     }
@@ -1186,12 +1255,13 @@ sc_toggle_index(girara_session_t* session, girara_argument_t* UNUSED(argument),
       goto error_free;
     }
 
-    document_index_build(model, NULL, document_index);
+    document_index_build(session, model, NULL, document_index);
     girara_node_free(document_index);
 
     /* setup widget */
     gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (treeview), 0, "Title", renderer, "markup", 0, NULL);
     gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (treeview), 1, "Target", renderer2, "text", 1, NULL);
+    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (treeview), 2, "(alt)", renderer2, "text", 2, NULL);
 
     gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
     g_object_set(G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
@@ -1362,7 +1432,7 @@ sc_toggle_presentation(girara_session_t* session, girara_argument_t*
     zathura->shortcut.toggle_presentation_mode.zoom = zathura_document_get_zoom(zathura->document);
 
     /* adjust window */
-    girara_argument_t argument = { ZATHURA_ADJUST_BESTFIT, NULL };
+    girara_argument_t argument = {.n = ZATHURA_ADJUST_BESTFIT, .data = NULL};
     sc_adjust_window(session, &argument, NULL, 0);
 
     /* hide status and inputbar */
@@ -1380,13 +1450,11 @@ sc_toggle_presentation(girara_session_t* session, girara_argument_t*
   return false;
 }
 
-bool
-sc_quit(girara_session_t* session, girara_argument_t* UNUSED(argument),
-        girara_event_t* UNUSED(event), unsigned int UNUSED(t))
-{
+bool sc_quit(girara_session_t* session, girara_argument_t* UNUSED(argument), girara_event_t* UNUSED(event),
+             unsigned int UNUSED(t)) {
   g_return_val_if_fail(session != NULL, false);
 
-  girara_argument_t arg = { GIRARA_HIDE, NULL };
+  girara_argument_t arg = {.n = GIRARA_HIDE, .data = NULL};
   girara_isc_completion(session, &arg, NULL, 0);
 
   cb_destroy(NULL, NULL);
@@ -1467,30 +1535,40 @@ sc_exec(girara_session_t* session, girara_argument_t* argument, girara_event_t*
     return false;
   }
 
+  girara_argument_t new_argument = *argument;
+  const char* bus_name           = zathura_dbus_get_name(zathura);
+  char* s                        = girara_replace_substring(new_argument.data, "$DBUS", bus_name);
+  if (s == NULL) {
+    return false;
+  }
+  new_argument.data = s;
+
   if (zathura->document != NULL) {
     const char* path = zathura_document_get_path(zathura->document);
+    unsigned int page = zathura_document_get_current_page_number(zathura->document);
+    char page_buf[G_ASCII_DTOSTR_BUF_SIZE];
+    g_ascii_dtostr(page_buf, G_ASCII_DTOSTR_BUF_SIZE, page+1);
 
-    girara_argument_t new_argument = *argument;
+    s = girara_replace_substring(new_argument.data, "$FILE", path);
+    g_free(new_argument.data);
 
-    char* r = girara_replace_substring(argument->data, "$FILE", path);
-    if (r == NULL) {
+    if (s == NULL) {
       return false;
     }
+    new_argument.data = s;
 
-    char* s = girara_replace_substring(r, "%", path);
-    g_free(r);
+    s = girara_replace_substring(new_argument.data, "$PAGE", page_buf);
+    g_free(new_argument.data);
 
     if (s == NULL) {
       return false;
     }
-
     new_argument.data = s;
-    const bool ret = girara_sc_exec(session, &new_argument, event, t);
-    g_free(s);
-    return ret;
   }
 
-  return girara_sc_exec(session, argument, event, t);
+  const bool ret = girara_sc_exec(session, &new_argument, event, t);
+  g_free(new_argument.data);
+  return ret;
 }
 
 bool
