@@ -28,10 +28,7 @@
 #include "callbacks.h"
 #include "config.h"
 #include "commands.h"
-#ifdef WITH_SQLITE
 #include "database-sqlite.h"
-#endif
-#include "database-plain.h"
 #include "document.h"
 #include "shortcuts.h"
 #include "zathura.h"
@@ -61,7 +58,6 @@ typedef struct zathura_document_info_s {
   char* search_string;
 } zathura_document_info_t;
 
-
 static gboolean document_info_open(gpointer data);
 
 #ifdef G_OS_UNIX
@@ -95,7 +91,10 @@ zathura_create(void)
 
   /* global settings */
   zathura->global.search_direction = FORWARD;
-  zathura->global.sandbox = ZATHURA_SANDBOX_NORMAL;
+  zathura->global.synctex_edit_modmask = GDK_CONTROL_MASK;
+  zathura->global.highlighter_modmask = GDK_SHIFT_MASK;
+  zathura->global.sandbox = ZATHURA_SANDBOX_NONE;
+  zathura->global.double_click_follow = true;
 
   /* plugins */
   zathura->plugins.manager = zathura_plugin_manager_new();
@@ -173,10 +172,12 @@ zathura_update_view_ppi(zathura_t* zathura)
   GdkRectangle monitor_geom;
   gdk_monitor_get_geometry(monitor, &monitor_geom);
 
-  /* calculate ppi, knowing that 1 inch = 25.4 mm */
-  if (width_mm == 0) {
+  /* Due to a bug in Gtk, width is sometimes incorrectly reported to be 1mm
+   * see https://gitlab.gnome.org/GNOME/gtk/issues/3115 for details */
+  if (width_mm <= 1) {
     girara_debug("cannot calculate PPI: monitor has zero width");
   } else {
+    /* calculate ppi, knowing that 1 inch = 25.4 mm */
     ppi = monitor_geom.width * 25.4 / width_mm;
   }
 
@@ -184,8 +185,7 @@ zathura_update_view_ppi(zathura_t* zathura)
   /* work around apparent bug in GDK: on Wayland, monitor geometry doesn't
    * return values in application pixels as documented, but in device pixels.
    * */
-  if (GDK_IS_WAYLAND_DISPLAY(display))
-  {
+  if (GDK_IS_WAYLAND_DISPLAY(display)) {
     /* not using the cached value for the scale factor here to avoid issues
      * if this function is called before the cached value is updated */
     const int device_factor = gtk_widget_get_scale_factor(zathura->ui.session->gtk.view);
@@ -206,9 +206,11 @@ zathura_update_view_ppi(zathura_t* zathura)
   }
 }
 
-static bool
-init_ui(zathura_t* zathura)
-{
+static void weak_ref_object_unref(void* data, GObject* UNUSED(object)) {
+  g_object_unref(data);
+}
+
+static bool init_ui(zathura_t* zathura) {
   if (girara_session_init(zathura->ui.session, "zathura") == false) {
     girara_error("Failed to initialize girara.");
     return false;
@@ -218,6 +220,14 @@ init_ui(zathura_t* zathura)
   zathura->ui.session->events.buffer_changed  = cb_buffer_changed;
   zathura->ui.session->events.unknown_command = cb_unknown_command;
 
+  /* gestures */
+
+  GtkGesture* zoom = gtk_gesture_zoom_new(GTK_WIDGET(zathura->ui.session->gtk.view));
+  g_signal_connect(zoom, "scale-changed", G_CALLBACK(cb_gesture_zoom_scale_changed), zathura);
+  g_signal_connect(zoom, "begin", G_CALLBACK(cb_gesture_zoom_begin), zathura);
+  gtk_event_controller_set_propagation_phase(GTK_EVENT_CONTROLLER(zoom), GTK_PHASE_BUBBLE);
+  g_object_weak_ref(G_OBJECT(zathura->ui.session->gtk.view), weak_ref_object_unref, zoom);
+
   /* zathura signals */
   zathura->signals.refresh_view = g_signal_new(
     "refresh-view", GTK_TYPE_WIDGET, G_SIGNAL_RUN_LAST, 0, NULL, NULL,
@@ -357,9 +367,7 @@ init_css(zathura_t* zathura)
   return true;
 }
 
-static void
-init_database(zathura_t* zathura)
-{
+static void init_database(zathura_t* zathura) {
   char* database = NULL;
   girara_setting_get(zathura->ui.session, "database", &database);
 
@@ -370,33 +378,27 @@ init_database(zathura_t* zathura)
 
   if (g_strcmp0(database, "plain") == 0) {
     girara_debug("Using plain database backend.");
-    zathura->database = zathura_plaindatabase_new(zathura->config.data_dir);
-#ifdef WITH_SQLITE
+    char* tmp         = g_build_filename(zathura->config.data_dir, "bookmarks.sqlite", NULL);
+    zathura->database = zathura_sqldatabase_new_from_plain(tmp, zathura->config.data_dir);
+    g_free(tmp);
   } else if (g_strcmp0(database, "sqlite") == 0) {
     girara_debug("Using sqlite database backend.");
-    char* tmp =
-      g_build_filename(zathura->config.data_dir, "bookmarks.sqlite", NULL);
+    char* tmp         = g_build_filename(zathura->config.data_dir, "bookmarks.sqlite", NULL);
     zathura->database = zathura_sqldatabase_new(tmp);
     g_free(tmp);
-#endif
   } else if (g_strcmp0(database, "null") != 0) {
     girara_error("Database backend '%s' is not supported.", database);
   }
 
   if (zathura->database == NULL && g_strcmp0(database, "null") != 0) {
-    girara_error(
-      "Unable to initialize database. Bookmarks won't be available.");
-  }
-  else {
-    g_object_set(G_OBJECT(zathura->ui.session->command_history), "io",
-                 zathura->database, NULL);
+    girara_error("Unable to initialize database. Bookmarks won't be available.");
+  } else {
+    g_object_set(G_OBJECT(zathura->ui.session->command_history), "io", zathura->database, NULL);
   }
   g_free(database);
 }
 
-static void
-init_jumplist(zathura_t* zathura)
-{
+static void init_jumplist(zathura_t* zathura) {
   int jumplist_size = 20;
   girara_setting_get(zathura->ui.session, "jumplist-size", &jumplist_size);
 
@@ -430,7 +432,9 @@ zathura_init(zathura_t* zathura)
   g_set_prgname("org.pwmt.zathura");
 
   /* load plugins */
-  zathura_plugin_manager_load(zathura->plugins.manager);
+  if (zathura_plugin_manager_load(zathura->plugins.manager) == false) {
+    girara_error("Found no plugins. Please install at least one plugin.");
+  }
 
   /* configuration */
   config_load_default(zathura);
@@ -442,16 +446,9 @@ zathura_init(zathura_t* zathura)
     case ZATHURA_SANDBOX_NONE:
       girara_debug("Sandbox deactivated.");
       break;
-    case ZATHURA_SANDBOX_NORMAL:
-      girara_debug("Basic sandbox allowing normal operation.");
-      if (seccomp_enable_basic_filter() != 0) {
-        girara_error("Failed to initialize basic seccomp filter.");
-        goto error_free;
-      }
-      break;
     case ZATHURA_SANDBOX_STRICT:
       girara_debug("Strict sandbox preventing write and network access.");
-      if (seccomp_enable_strict_filter() != 0) {
+      if (seccomp_enable_strict_filter(zathura) != 0) {
         girara_error("Failed to initialize strict seccomp filter.");
         goto error_free;
       }
@@ -468,7 +465,7 @@ zathura_init(zathura_t* zathura)
   }
 
   /* disable unsupported features in strict sandbox mode */
-  if (zathura->global.sandbox != ZATHURA_SANDBOX_STRICT){
+  if (zathura->global.sandbox != ZATHURA_SANDBOX_STRICT) {
     /* database */
     init_database(zathura);
   }
@@ -746,7 +743,6 @@ prepare_document_open_from_stdin(const char* path)
   if (input_stream == NULL) {
     girara_error("Can not read from file descriptor.");
     return NULL;
-
   }
 
   GFileIOStream* iostream = NULL;
@@ -847,8 +843,7 @@ document_info_open(gpointer data)
       if (g_file_is_native(gf) == TRUE) {
         /* file was given as a native path */
         file = g_file_get_path(gf);
-      }
-      else {
+      } else {
         /* copy file with GIO */
         uri = g_file_get_uri(gf);
         file = prepare_document_open_from_gfile(gf);
@@ -868,7 +863,7 @@ document_info_open(gpointer data)
                               document_info->password, document_info->synctex);
       } else {
         document_open(document_info->zathura, file, uri, document_info->password,
-                      document_info->page_number);
+                      document_info->page_number, NULL);
       }
       g_free(file);
       g_free(uri);
@@ -963,9 +958,57 @@ document_open_password_dialog(gpointer data)
   return FALSE;
 }
 
+typedef struct {
+  double h;
+  double w;
+  int freq;
+} sample_t;
+
+static int document_page_size_comp(const void* a, const void* b) {
+  const sample_t* lhs = a;
+  const sample_t* rhs = b;
+  return rhs->freq - lhs->freq;
+}
+
+static void document_open_page_most_frequent_size(zathura_document_t* document, unsigned int* width,
+                                                  unsigned int* height) {
+  girara_list_t* samples             = girara_list_new2(g_free);
+  const unsigned int number_of_pages = zathura_document_get_number_of_pages(document);
+
+  for (unsigned int page_id = 0; page_id < number_of_pages; ++page_id) {
+    zathura_page_t* page = zathura_document_get_page(document, page_id);
+    const double w       = zathura_page_get_width(page);
+    const double h       = zathura_page_get_height(page);
+
+    bool found = false;
+    for (size_t idx = 0; idx != girara_list_size(samples) && !found; ++idx) {
+      sample_t* sample = girara_list_nth(samples, idx);
+      if (fabs(sample->h - h) <= DBL_EPSILON && fabs(sample->w - w) <= DBL_EPSILON) {
+        sample->freq++;
+      }
+    }
+
+    if (found == false) {
+      sample_t* sample = g_try_malloc0(sizeof(sample_t));
+      sample->w        = w;
+      sample->h        = h;
+      sample->freq     = 1;
+      girara_list_append(samples, sample);
+    }
+  }
+
+  girara_list_sort(samples, document_page_size_comp);
+
+  sample_t* max_sample = girara_list_nth(samples, 0);
+  *width               = max_sample->w;
+  *height              = max_sample->h;
+
+  girara_list_free(samples);
+}
+
 bool
 document_open(zathura_t* zathura, const char* path, const char* uri, const char* password,
-              int page_number)
+              int page_number, zathura_fileinfo_t* file_info_p)
 {
   if (zathura == NULL || zathura->plugins.manager == NULL || path == NULL) {
     goto error_out;
@@ -1025,8 +1068,13 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     .position_x = 0,
     .position_y = 0
   };
+
   bool known_file = false;
-  if (zathura->database != NULL) {
+  if( file_info_p ) {
+    file_info =  *file_info_p;
+    known_file = true;
+  }
+  else if (zathura->database != NULL) {
     const uint8_t* file_hash = zathura_document_get_hash(document);
     known_file = zathura_db_get_fileinfo(zathura->database, file_path, file_hash, &file_info);
   }
@@ -1165,12 +1213,22 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     goto error_free;
   }
 
+  unsigned int most_freq_width, most_freq_height;
+  document_open_page_most_frequent_size(document, &most_freq_width,
+                                        &most_freq_height);
+
   for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
     zathura_page_t* page = zathura_document_get_page(document, page_id);
     if (page == NULL) {
       goto error_free;
     }
 
+    unsigned int cell_height = 0, cell_width = 0;
+    zathura_document_get_cell_size(document, &cell_height, &cell_width);
+    zathura_document_set_cell_size(document, most_freq_height, most_freq_width);
+    zathura_page_set_width(page, most_freq_width);
+    zathura_page_set_height(page, most_freq_height);
+
     GtkWidget* page_widget = zathura_page_widget_new(zathura, page);
     if (page_widget == NULL) {
       goto error_free;
@@ -1277,6 +1335,9 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     position_set(zathura, file_info.position_x, file_info.position_y);
   }
 
+  bool show_signature_information = false;
+  girara_setting_get(zathura->ui.session, "show-signature-information", &show_signature_information);
+  zathura_show_signature_information(zathura, show_signature_information);
   update_visible_pages(zathura);
 
   return true;
@@ -1295,7 +1356,7 @@ document_open_synctex(zathura_t* zathura, const char* path, const char* uri,
                       const char* password, const char* synctex)
 {
   bool ret = document_open(zathura, path, password, uri,
-                           ZATHURA_PAGE_NUMBER_UNSPECIFIED);
+                           ZATHURA_PAGE_NUMBER_UNSPECIFIED, NULL);
   if (ret == false) {
     return false;
   }
@@ -1391,11 +1452,10 @@ remove_page_from_table(GtkWidget* page, gpointer UNUSED(permanent))
   gtk_container_remove(GTK_CONTAINER(gtk_widget_get_parent(page)), page);
 }
 
-static void
-save_fileinfo_to_db(zathura_t* zathura)
+zathura_fileinfo_t
+zathura_get_fileinfo(zathura_t* zathura)
 {
-  const char* path = zathura_document_get_path(zathura->document);
-  const uint8_t* file_hash = zathura_document_get_hash(zathura->document);
+  /* Caller needs to g_free(file_info.first_page_column_list) */
 
   zathura_fileinfo_t file_info = {
     .current_page = zathura_document_get_current_page_number(zathura->document),
@@ -1416,6 +1476,17 @@ save_fileinfo_to_db(zathura_t* zathura)
   girara_setting_get(zathura->ui.session, "page-right-to-left",
                      &(file_info.page_right_to_left));
 
+  return file_info;
+}
+
+static void
+save_fileinfo_to_db(zathura_t* zathura)
+{
+  const char* path = zathura_document_get_path(zathura->document);
+  const uint8_t* file_hash = zathura_document_get_hash(zathura->document);
+
+  zathura_fileinfo_t file_info = zathura_get_fileinfo(zathura);
+
   /* save file info */
   zathura_db_set_fileinfo(zathura->database, path, file_hash, &file_info);
   /* save jumplist */
@@ -1424,9 +1495,30 @@ save_fileinfo_to_db(zathura_t* zathura)
   g_free(file_info.first_page_column_list);
 }
 
-bool
-document_close(zathura_t* zathura, bool keep_monitor)
-{
+bool document_predecessor_free(zathura_t* zathura) {
+  if (zathura == NULL || (zathura->predecessor_document == NULL && zathura->predecessor_pages == NULL)) {
+    return false;
+  }
+
+  if (zathura->predecessor_pages != NULL) {
+    for (unsigned int i = 0; i < zathura_document_get_number_of_pages(zathura->predecessor_document); i++) {
+      g_object_unref(zathura->predecessor_pages[i]);
+    }
+    free(zathura->predecessor_pages);
+    zathura->predecessor_pages = NULL;
+    girara_debug("freed predecessor pages");
+  }
+  if (zathura->predecessor_document != NULL) {
+    /* remove document */
+    zathura_document_free(zathura->predecessor_document);
+    zathura->predecessor_document = NULL;
+    girara_debug("freed predecessor document");
+  }
+
+  return true;
+}
+
+bool document_close(zathura_t* zathura, bool keep_monitor) {
   if (zathura == NULL || zathura->document == NULL) {
     return false;
   }
@@ -1474,17 +1566,43 @@ document_close(zathura_t* zathura, bool keep_monitor)
   /* release render thread */
   g_clear_object(&zathura->sync.render_thread);
 
+  /* keep the current state to prevent flicker? */
+  bool override_predecessor = keep_monitor;
+
+  if (override_predecessor) {
+    /* do not override predecessor buffer with empty pages */
+    unsigned int cur_page_num = zathura_document_get_current_page_number(zathura->document);
+    ZathuraPage* cur_page     = ZATHURA_PAGE(zathura->pages[cur_page_num]);
+    if (!zathura_page_widget_have_surface(cur_page)) {
+      override_predecessor = false;
+    }
+  }
+
+  /* free predecessor buffer if we want to overwrite it or if we destroy the document for good */
+  if (override_predecessor || !keep_monitor) {
+    document_predecessor_free(zathura);
+  }
+
   /* remove widgets */
   gtk_container_foreach(GTK_CONTAINER(zathura->ui.page_widget), remove_page_from_table, NULL);
-  for (unsigned int i = 0; i < zathura_document_get_number_of_pages(zathura->document); i++) {
-    g_object_unref(zathura->pages[i]);
-  }
-  free(zathura->pages);
-  zathura->pages = NULL;
 
-  /* remove document */
-  zathura_document_free(zathura->document);
-  zathura->document = NULL;
+  if (!override_predecessor) {
+    for (unsigned int i = 0; i < zathura_document_get_number_of_pages(zathura->document); i++) {
+      g_object_unref(zathura->pages[i]);
+    }
+    free(zathura->pages);
+    zathura->pages = NULL;
+
+    /* remove document */
+    zathura_document_free(zathura->document);
+    zathura->document = NULL;
+  } else {
+    girara_debug("preserving pages and document as predecessor");
+    zathura->predecessor_pages    = zathura->pages;
+    zathura->pages                = NULL;
+    zathura->predecessor_document = zathura->document;
+    zathura->document             = NULL;
+  }
 
   /* remove index */
   if (zathura->ui.index != NULL) {
@@ -1542,17 +1660,29 @@ statusbar_page_number_update(zathura_t* zathura)
 
   unsigned int number_of_pages     = zathura_document_get_number_of_pages(zathura->document);
   unsigned int current_page_number = zathura_document_get_current_page_number(zathura->document);
+  unsigned int page_number_percent = number_of_pages ? 100 * (current_page_number + 1) / number_of_pages : 0;
 
   if (zathura->document != NULL) {
     zathura_page_t* page = zathura_document_get_page(zathura->document, current_page_number);
     char* page_label = zathura_page_get_label(page, NULL);
 
+    bool show_percent = false;
+    girara_setting_get(zathura->ui.session, "statusbar-page-percent", &show_percent);
+
     char* page_number_text = NULL;
     if (page_label != NULL) {
-      page_number_text = g_strdup_printf("[%s (%d/%d)]", page_label, current_page_number + 1, number_of_pages);
+      if (show_percent) {
+        page_number_text = g_strdup_printf("[%s (%d/%d) (%d%%)]", page_label, current_page_number + 1, number_of_pages, page_number_percent);
+      } else {
+        page_number_text = g_strdup_printf("[%s (%d/%d)]", page_label, current_page_number + 1, number_of_pages);
+      }
       g_free(page_label);
     } else {
-      page_number_text = g_strdup_printf("[%d/%d]", current_page_number + 1, number_of_pages);
+      if (show_percent) {
+        page_number_text = g_strdup_printf("[%d/%d (%d%%)]", current_page_number + 1, number_of_pages, page_number_percent);
+      } else {
+        page_number_text = g_strdup_printf("[%d/%d]", current_page_number + 1, number_of_pages);
+      }
     }
     girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.page_number, page_number_text);
 
@@ -1757,3 +1887,21 @@ zathura_signal_sigterm(gpointer data)
   return TRUE;
 }
 #endif
+
+void zathura_show_signature_information(zathura_t* zathura, bool show) {
+  if (zathura->document == NULL) {
+    return;
+  }
+
+  GValue show_sig_info_value = {0};
+  g_value_init(&show_sig_info_value, G_TYPE_BOOLEAN);
+  g_value_set_boolean(&show_sig_info_value, show);
+
+  const unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
+  for (unsigned int page = 0; page != number_of_pages; ++page) {
+    // draw signature info
+    g_object_set_property(G_OBJECT(zathura->pages[page]), "draw-signatures", &show_sig_info_value);
+  }
+
+  g_value_unset(&show_sig_info_value);
+}
